/*
 * rbcd.h
 *
 *   Copyright (C) Ueno Katsuhiro 2000
 *
 * $Id: rbcd.h,v 1.1.1.1 2000/10/15 06:13:34 katsu Exp $
 */

#ifndef RUBY__RBCD_H
#define RUBY__RBCD_H


extern VALUE rb_eCDROMNoDisc;


struct rbcd_time {
	int lba;                /* use msf if lba < 0 */
	struct rbcd_time_msf {
		int min;
		int sec;
		int frame;
	} msf;
};

#define RBCD_MSF_OFFSET  150
#define RBCD_MSF_FRAMES  75

struct rbcd_track {
	int track_no;
	enum { RBCD_LEADOUT, RBCD_AUDIO_TRACK, RBCD_DATA_TRACK } track_type;
	struct rbcd_time start_addr;
};

#define RBCD_TRACK_TYPE_STR  "leadout", "audio", "data"

struct rbcd_status {
	enum { RBCD_PLAYING, RBCD_PAUSED, RBCD_COMPLETED, RBCD_NO_STATUS
	} status;
	int track_no;
	struct rbcd_time addr;
	struct rbcd_time_msf track_time;
};

#define RBCD_STATUS_STR  "playing", "paused", "completed", "no status"

union rbcd_device {
	int fd;
	void *p;
};

void rbcd_open(union rbcd_device *cdrom, const char *device);
void rbcd_close(union rbcd_device cdrom);
void rbcd_inject(union rbcd_device cdrom);
void rbcd_eject(union rbcd_device cdrom);
void rbcd_lock_door(union rbcd_device cdrom, int locked);

void rbcd_get_track_range(union rbcd_device cdrom, int *start, int *end);
void rbcd_get_track(union rbcd_device cdrom,
		    int track_no, struct rbcd_track *track);
int rbcd_get_last_session_addr(union rbcd_device cdrom,
			       struct rbcd_time *addr);

void rbcd_play_audio(union rbcd_device cdrom,
		     struct rbcd_time *start, struct rbcd_time *end);
void rbcd_pause_audio(union rbcd_device cdrom);
void rbcd_resume_audio(union rbcd_device cdrom);
void rbcd_stop_audio(union rbcd_device cdrom);

#define RBCD_FRAMESIZE_MODE1  2048
#define RBCD_FRAMESIZE_MODE2  2336
#define RBCD_FRAMESIZE_RAW    2352
#define RBCD_FRAMESIZE_AUDIO  2352

void rbcd_read_audio(union rbcd_device cdrom,
		     struct rbcd_time *addr, int frames, char *dst);
void rbcd_read_mode1(union rbcd_device cdrom,
		     struct rbcd_time *addr, int frames, char *dst);
void rbcd_read_mode2(union rbcd_device cdrom,
		     struct rbcd_time *addr, int frames, char *dst);
void rbcd_read_raw(union rbcd_device cdrom,
		   struct rbcd_time *addr, int frames, char *dst);

void rbcd_get_status(union rbcd_device cdrom, struct rbcd_status *status);
void rbcd_get_volume(union rbcd_device cdrom, int *left, int *right);
void rbcd_set_volume(union rbcd_device cdrom, int left, int right);

int rbcd_num_slots(union rbcd_device cdrom);
int rbcd_change_slot(union rbcd_device cdrom, int slot);
int rbcd_is_ready(union rbcd_device cdrom, int slot);

#endif  /* RUBY__RBCD_H */
