#
# rbcd-gtk.rb
#
#   Copyright (C) Ueno Katsuhiro 2000
#
# $Id: rbcd-gtk.rb,v 1.2 2000/10/20 19:06:13 katsu Exp $
#

require 'gtk'
require 'rbcd'


module Gtk

  class CList

    unless instance_methods.include? 'get_selection_info' then
      def get_selection_info(x, y)
        adj = vadjustment
        if y < 0 or y >= adj.page_size then
          nil
        else
          row = (y + adj.value.to_i) / (row_height + 1)
          if row < rows then
            [ 0, row ]
          else
            nil
          end
        end
      end
    end

    unless instance_methods.include? 'find_row_from_data' then
      def find_row_from_data(data)
        rows.times { |row|
          return row if get_row_data(row) == data
        }
        nil
      end
    end

  end

end




class ResistantAdjustment < Gtk::Adjustment

  alias __orig_get_value  value
  alias __orig_set_value  set_value

  def initialize(resist_point, resistance, value, lower, upper,
                 step_inc, page_inc, page_size)
    super value, lower, upper + resistance, step_inc, page_inc, page_size
    @resist_point, @resistance = resist_point, resistance
    signal_connect('value_changed') {
      v = __orig_get_value
      if v >= @resist_point and v <= @resist_point + @resistance then
        __orig_set_value @resist_point + @resistance / 2
      end
    }
  end

  def value
    v = __orig_get_value
    if v < @resist_point then
      v
    elsif v <= @resist_point + @resistance then
      @resist_point
    else
      v - @resistance
    end
  end

  def set_value(v)
    if v < @resist_point then
      __orig_set_value v
    elsif v == @resist_point then
      __orig_set_value @resist_point + @resistance / 2
    else
      __orig_set_value v + @resistance
    end
  end

  alias value=  set_value

end




class RbcdGtk

  PrevXpm = [
    '11 11 6 1',
    '  c None',    '. c #000000', '+ c #D1D1D1',
    '# c #A3A3A3', '@ c #939393', '$ c #707070',
    '....     ..',
    '.+@.    .#.',
    '.+$.   .+$.',
    '.+$.  .+@$.',
    '.+$. .+@@$.',
    '.+$..+@@@$.',
    '.+$. .#@@$.',
    '.+$.  .#@$.',
    '.+$.   .#$.',
    '.@$.    .$.',
    '....     ..'
  ]
  NextXpm = [
    '11 11 6 1',
    '  c None',    '. c #000000', '+ c #D1D1D1',
    '# c #A3A3A3', '@ c #939393', '$ c #707070',
    '..     ....',
    '.+.    .+@.',
    '.+#.   .+$.',
    '.+@#.  .+$.',
    '.+@@#. .+$.',
    '.+@@@$..+$.',
    '.+@@$. .+$.',
    '.+@$.  .+$.',
    '.+$.   .+$.',
    '.+.    .@$.',
    '..     ....'
  ]
  PlayXpm = [
    '11 11 6 1',
    '  c None',    '. c #000000', '+ c #D1D1D1',
    '# c #A3A3A3', '@ c #939393', '$ c #707070',
    '..         ',
    '.+..       ',
    '.+#+..     ',
    '.+@@#+..   ',
    '.+@@@@#+.. ',
    '.+@@@@@@$$.',
    '.+@@@@$$.. ',
    '.+@@$$..   ',
    '.+$$..     ',
    '.#..       ',
    '..         '
  ]
  PauseXpm = [
    '10 11 5 1',
    '  c None',    '. c #000000', '+ c #D1D1D1',
    '@ c #939393', '$ c #707070',
    '....  ....',
    '.+@.  .+@.',
    '.+$.  .+$.',
    '.+$.  .+$.',
    '.+$.  .+$.',
    '.+$.  .+$.',
    '.+$.  .+$.',
    '.+$.  .+$.',
    '.+$.  .+$.',
    '.@$.  .@$.',
    '....  ....'
  ]
  StopXpm = [
    '10 11 4 1',
    '. c #000000', '+ c #D1D1D1', '@ c #939393', '$ c #707070',
    '..........',
    '.+++++++@.',
    '.+@@@@@@$.',
    '.+@@@@@@$.',
    '.+@@@@@@$.',
    '.+@@@@@@$.',
    '.+@@@@@@$.',
    '.+@@@@@@$.',
    '.+@@@@@@$.',
    '.@$$$$$$$.',
    '..........'
  ]
  EjectXpm = [
    '11 11 6 1',
    '  c None',    '. c #000000', '+ c #D1D1D1',
    '# c #A3A3A3', '@ c #939393', '$ c #707070',
    '     .     ',
    '    .+.    ',
    '   .+@#.   ',
    '  .+@@@$.  ',
    ' .+@@@@@$. ',
    '.#$$$$$$$$.',
    '...........',
    '           ',
    '...........',
    '.+###@@@@$.',
    '...........'
  ]
  RepeatXpm = [
    '11 11 6 1',
    '  c None',    '. c #000000', '+ c #D1D1D1',
    '# c #A3A3A3', '@ c #939393', '$ c #707070',
    ' ........  ',
    ' .+###@@$. ',
    ' .......#$.',
    '       $.#.',
    '   .    .#.',
    '  ..   $.#.',
    ' .#.....#$.',
    '.####@@@$. ',
    ' .@......  ',
    '  ..       ',
    '   .       '
  ]
  VolumeXpm = [
    '16 9 4 1',
    '  c None', '. c #000000', '% c #F0F0F0', '+ c #D1D1D1',
    '     ..      .  ',
    '    .%.       . ',
    '....%%.   .   . ',
    '.%.%%+.    .   .',
    '.%.%%+.    .   .',
    '.+.%++.    .   .',
    '....++.   .   . ',
    '    .+.       . ',
    '     ..      .  ',
  ]
  BalanceXpm = [
    '27 9 4 1',
    '  c None', '. c #000000', '% c #F0F0F0', '+ c #D1D1D1',
    '  .                     .  ',
    ' .                       . ',
    ' .   .   .........   .   . ',
    '.   .    .%%%%%++.    .   .',
    '.   .     .%%%++.     .   .',
    '.   .      .%%+.      .   .',
    ' .   .     .....     .   . ',
    ' .         .%%+.         . ',
    '  .        .....        .  ',
  ]


  RCFile = ENV['HOME'] + File::Separator + '.rbcdrc'


  private

  def create_pixmap(data)
    p, m = Gdk::Pixmap.create_from_xpm_d(@window.window,
                                         @window.style.bg(Gtk::STATE_NORMAL),
                                         data);
    Gtk::Pixmap.new(p, m)
  end

  def create_pixmap_button(data, klass = Gtk::Button)
    pixmap = create_pixmap(data)
    button = klass.new
    button.add pixmap
    pixmap.show
    button
  end

  def create_box(klass, children)
    klass = [ klass, false, 0 ] unless klass.is_a? Array
    box = klass[0].new(*klass[1, klass.size])
    children.each { |w,expand,fill,padding|
      expand = false if expand.nil?
      fill = false if fill.nil?
      padding = 0 if padding.nil?
      box.pack_start w, expand, fill, padding
      w.show
    }
    box
  end

  def pack_start_and_show(box, w, expand, fill, padding)
    box.pack_start w, expand, fill, padding
    w.show
  end

  def pack_end_and_show(box, w, expand, fill, padding)
    box.pack_end w, expand, fill, padding
    w.show
  end


  ToolBox = [
    [ 'on_prev',  PrevXpm,  'Previous Track', ],
    [ 'on_pause', PauseXpm, 'Pause', ],
    [ 'on_play',  PlayXpm,  'Play', ],
    [ 'on_stop',  StopXpm,  'Stop', ],
    [ 'on_next',  NextXpm,  'Next Track', ],
  ]

  BalanceRegistance = 30.0


  def initialize(file)
    Gtk::RC.parse RCFile

    @editor = nil
    @window = Gtk::Window.new(Gtk::WINDOW_TOPLEVEL)
    @window.set_name Rbcd::PlayerName
    @window.set_title Rbcd::PlayerName
    @window.signal_connect('destroy') { Gtk.main_quit }
    @window.realize

    def @window.set_title(title)
      if title.empty? then
        super Rbcd::PlayerName
      else
        super sprintf('%s - %s', Rbcd::PlayerName, title)
      end
    end

    tooltips = Gtk::Tooltips.new
    tooltips.set_delay 1500

    @toolbox = Gtk::HBox.new(true, 0)
    ToolBox.each { |action,xpm,tips|
      button = create_pixmap_button(xpm)
      button.set_usize 28, 23
      button.set_relief Gtk::RELIEF_NONE
      button.signal_connect('clicked', &eval("proc { #{action}; true }"))
      tooltips.set_tip button, tips, nil
      pack_start_and_show @toolbox, button, true, true, 0
    }

    repeat_button = Gtk::ToggleButton.new('R')
    repeat_button = create_pixmap_button(RepeatXpm, Gtk::ToggleButton)
    repeat_button.set_usize 28, 23
    repeat_button.set_relief Gtk::RELIEF_NONE
    tooltips.set_tip repeat_button, 'Repeat All Tracks', nil

    repeat_button.signal_connect('toggled') {
      @rbcd.repeat = repeat_button.active?
      true
    }

    eject_button = create_pixmap_button(EjectXpm)
    eject_button.set_usize 28, 23
    eject_button.set_relief Gtk::RELIEF_NONE
    tooltips.set_tip eject_button, 'Eject', nil

    eject_button.signal_connect('clicked') { on_eject; true }

    @title_label = Gtk::Label.new('Title / Artist')
    @track_label = Gtk::Label.new('00/00')
    @time_now_label = Gtk::Label.new('00:00')
    @time_max_label = Gtk::Label.new(' / 00:00')
    @title_label.set_name 'title_label'
    @track_label.set_name 'track_label'

    def @title_label.set_title(title, artist)
      title = '(no title)' if title.empty?
      title += ' / ' + artist unless artist.empty?
      set_text title
    end
    def @title_label.clear
      set_text 'no disc'
    end

    def @track_label.set_track(m, n)
      set_text sprintf('%02d/%02d', m, n)
    end
    def @track_label.clear
      set_text '00/00'
    end

    def @time_now_label.set_time(t)
      set_text sprintf('%02d:%02d', t.min, t.sec)
    end
    def @time_now_label.clear
      set_text '00:00'
    end

    def @time_max_label.set_time(t)
      set_text sprintf(' / %02d:%02d', t.min, t.sec)
    end
    def @time_max_label.clear
      set_text ' / 00:00'
    end

    track_box = Gtk::HBox.new(false, 0)
    pack_start_and_show track_box, @time_now_label, false, false, 0
    pack_start_and_show track_box, @time_max_label, false, false, 0

    @display = Gtk::Frame.new(nil)
    @display.set_shadow_type Gtk::SHADOW_ETCHED_IN
    eventbox = Gtk::EventBox.new
    eventbox.set_name 'display'
    eventbox.set_events Gdk::BUTTON_PRESS_MASK
    hbox = Gtk::HBox.new(false, 0)
    pack_start_and_show hbox, @track_label, true, true, 8
    pack_start_and_show hbox, track_box, true, false, 8
    eventbox.add hbox
    hbox.show
    @display.add eventbox
    eventbox.show

    @display_abs_time = false
    eventbox.signal_connect('button_press_event') {
      toggle_display_abs_time
      false
    }

    @play_scale = Gtk::HScale.new(nil)
    @play_scale.set_name 'play_scale'
    @play_scale.set_draw_value false

    @play_scale_changing = false
    @play_scale.signal_connect('button_press_event') {
      @play_scale_changing = true
    }
    @play_scale.signal_connect('button_release_event') {
      @play_scale_changing = false
      play_scale_changed
      true
    }
    @play_scale.signal_connect('key_press_event') {
      play_scale_changed
      true
    }

    @volume_adj = Gtk::Adjustment.new(0.0, 0.0, 255.0, 1.0, 25.5, 0.0)
    @volume_adj.signal_connect('value_changed') { on_change_volume; false }
    volume_scale = Gtk::HScale.new(@volume_adj)
    volume_scale.set_draw_value false
    tooltips.set_tip volume_scale, 'Volume', nil

    @balance_adj = ResistantAdjustment.new(50.0, BalanceRegistance,
                                           0.0, 0.0, 100.0, 1.0,
                                           (BalanceRegistance/20).ceil * 10,
                                           0.0)
    @balance_adj.signal_connect('value_changed') { on_change_volume; false }
    balance_scale = Gtk::HScale.new(@balance_adj)
    balance_scale.set_draw_value false
    tooltips.set_tip balance_scale, 'Balance', nil

    volumebox = Gtk::HBox.new(false, 10)
    hbox = Gtk::HBox.new(false, 2)
    pack_start_and_show hbox, create_pixmap(VolumeXpm), false, false, 0
    pack_start_and_show hbox, volume_scale, true, true, 0
    pack_start_and_show volumebox, hbox, true, true, 0
    hbox = Gtk::HBox.new(false, 2)
    pack_start_and_show hbox, create_pixmap(BalanceXpm), false, false, 0
    pack_start_and_show hbox, balance_scale, true, true, 0
    pack_start_and_show volumebox, hbox, true, true, 0

    subtools = Gtk::HBox.new(false, 0)
    pack_start_and_show subtools, repeat_button, false, false, 0
    pack_start_and_show subtools, @display, true, true, 4
    pack_start_and_show subtools, eject_button, false, false, 0

    control = Gtk::VBox.new(false, 2)
    pack_start_and_show control, @play_scale, false, false, 0
    hbox = Gtk::HBox.new(false, 2)
    pack_start_and_show hbox, @toolbox, false, false, 0
    vbox = Gtk::VBox.new(false, 1)
    pack_start_and_show vbox, volumebox, true, true, 0
    pack_start_and_show vbox, subtools, true, true, 0
    pack_start_and_show hbox, vbox, true, true, 0
    pack_start_and_show control, hbox, false, false, 0

    scroll = Gtk::ScrolledWindow.new
    scroll.set_policy Gtk::POLICY_NEVER, Gtk::POLICY_AUTOMATIC

    @playlist = Gtk::CList.new(['Slot #', 'Track #', 'Time', 'Title'])
    @playlist.set_name 'playlist'
    @playlist.set_column_width 0, 20
    @playlist.set_column_width 1, 20
    @playlist.set_column_width 2, 60
    @playlist.column_titles_passive
    @playlist.column_titles_hide
    @playlist.set_selection_mode Gtk::SELECTION_SINGLE
    @playlist.set_reorderable true
    scroll.add @playlist
    @playlist.show

    @current = nil
    @playlist.signal_connect('select_row') { |w,row,col,event|
      w.unselect_row row, 0 unless row == @current
      false
    }
    @playlist.signal_connect('unselect_row') { |w,row,col,event|
      w.select_row @current, 0 if @current
      false
    }
    @playlist.signal_connect('row_move') { |w,from,to|
      to += 1 if from < to
      move_track w.get_row_data(from), w.get_row_data(to)
      @current = to if @current == from
      true
    }
    @playlist.signal_connect('event') { |w,event|
      case event.event_type
      when Gdk::BUTTON_PRESS then
        col, row = w.get_selection_info(event.x, event.y)
        case event.button
        when 2 then
          if row then
            track = w.get_row_data(row)
            toggle_track_enable track
            w.set_track_style row, track
          end
          true
        when 3 then
          if row then
            w.select_row row, 0
            popup_menu event, w.get_row_data(row)
          end
          true
        end
      when Gdk::BUTTON2_PRESS then
        col, row = w.get_selection_info(event.x, event.y)
        if row then
          track = w.get_row_data(row)
          case event.button
          when 1 then
            track.play if row
            true
          when 2 then
            toggle_all_tracks_enable track
            true
          end
        end
      else
        false
      end
    }

    def @playlist.set_track_style(row, track)
      s = style
      unless track.enable? then
        s = s.copy
        s.set_fg Gtk::STATE_NORMAL, 40960, 40960, 40960
        s.set_fg Gtk::STATE_SELECTED, 49152, 49152, 49152
      end
      set_row_style row, s
    end

    vbox = Gtk::VBox.new(false, 4)
    pack_start_and_show vbox, scroll, true, true, 0
    pack_start_and_show vbox, @title_label, false, false, 0
    pack_start_and_show vbox, control, false, false, 0

    @window.border_width = 4
    @window.add vbox
    vbox.show

    @window.ensure_style
    scroll.set_usize(-1, @playlist.row_height * 4 + 10)
    font = @playlist.get_style.font
    @playlist.set_column_width 0, font.string_width('0')
    @playlist.set_column_width 1, font.string_width('00 ')
    @playlist.set_column_width 2, font.string_width('00:00.00 ')

    @rbcd = Rbcd.new(file, self)
    volume, balance = @rbcd.get_volume
    @balance_adj.value = balance
    @volume_adj.value = volume

    @title_label.clear
    @play_scale.set_sensitive false
    @toolbox.set_sensitive false

    @window.show
  end


  public

  def main
    Gtk.timeout_add(250) {
      @rbcd.dispatch
      true
    }
    Gtk.main
  end


  private

  def on_prev
    @rbcd.prev_track
  end

  def on_next
    @rbcd.next_track
  end

  def on_stop
    @rbcd.stop
  end

  def on_pause
    @rbcd.pause
  end

  def on_play
    @rbcd.play
  end

  def on_eject
    @rbcd.eject
  end

  def on_change_volume
    @rbcd.set_volume @volume_adj.value, @balance_adj.value
  end

  def play_scale_changed
    @rbcd.play_from @play_scale.get_adjustment.value.to_i
  end

  def toggle_display_abs_time
    @display_abs_time = (not @display_abs_time)
    if @current then
      track = @playlist.get_row_data(@current)
    elsif @playlist.rows > 0 then
      track = @playlist.get_row_data(0)
    else
      return
    end
    if @display_abs_time then
      t = track.disc.size
    else
      t = track.size
    end
    @time_max_label.set_time t
  end

  def toggle_track_enable(track)
    track.enable = (not track.enable)
    @rbcd.save_disc track.disc
  end

  def toggle_all_tracks_enable(track)
    f = (not track.enable)
    @playlist.rows.times { |row|
      track = @playlist.get_row_data(row)
      track.enable = f
      @playlist.set_track_style row, track
    }
    @rbcd.save_all_discs
  end

  def move_track(track, track_to)
    @rbcd.move_track track, track_to
    @rbcd.save_disc track.disc
  end


  def sort_playlist
    @rbcd.sort_tracks
    @rbcd.save_all_discs
  end


  def popup_menu(event, track)
    menu = Gtk::Menu.new
    edit_item = Gtk::MenuItem.new('Edit disc infomations')
    sort_item = Gtk::MenuItem.new('Sort all tracks')
    sep1 = Gtk::MenuItem.new
    sep2 = Gtk::MenuItem.new

    if @playlist.rows == 0 then
      edit_item.set_sensitive false
      sort_item.set_sensitive false
    end
    menu.append sort_item
    menu.append sep2
    menu.append edit_item
    edit_item.show
    sort_item.show
    sep1.show
    sep2.show

    edit_item.signal_connect('activate') { open_editor track; true }
    sort_item.signal_connect('activate') { sort_playlist; true }
    menu.popup nil, nil, nil, event.button, event.time
  end


  def open_editor(track)
    disc = track.disc
    if @editor and @editor[disc] then
      @editor[disc].window.raise
      @editor[disc].grab_focus
      return
    end
    disc_id = @rbcd.disc_id(disc)
    window = Gtk::Window.new(Gtk::WINDOW_TOPLEVEL)
    window.set_title Rbcd::PlayerName + ' disc editor - ' + disc_id
    window.border_width = 4
    window.signal_connect('destroy') { @editor.delete disc; false }
    @editor = {} unless @editor
    @editor[disc] = window

    title_box = Gtk::Table.new(3, 2)
    title_box.set_col_spacings 4
    title_box.set_row_spacings 4

    label = Gtk::Label.new('Disc ID:')
    label.show
    title_box.attach label, 0, 1, 0, 1, Gtk::FILL, 0, 0
    entry = Gtk::Entry.new
    entry.set_text disc_id
    entry.set_sensitive false
    entry.set_editable false
    entry.show
    title_box.attach entry, 1, 2, 0, 1, Gtk::FILL | Gtk::EXPAND, 0, 0

    title_entry = Gtk::Entry.new
    artist_entry = Gtk::Entry.new
    label = Gtk::Label.new('Disc Title:')
    label.show
    title_box.attach label, 0, 1, 1, 2, Gtk::FILL, 0, 0
    title_box.attach title_entry, 1, 2, 1, 2, Gtk::FILL | Gtk::EXPAND, 0, 0
    label = Gtk::Label.new('Artist:')
    label.show
    title_box.attach label, 0, 1, 2, 3, Gtk::FILL, 0, 0
    title_box.attach artist_entry, 1, 2, 2, 3, Gtk::FILL | Gtk::EXPAND, 0, 0
    title_entry.show
    artist_entry.show
    title_entry.set_text disc.title
    artist_entry.set_text disc.artist

    track_entries = []
    tracks = disc.audio_tracks
    table = Gtk::Table.new(tracks.size, 2)
    table.border_width = 4
    table.set_col_spacings 4
    table.set_row_spacings 4
    tracks.each_with_index { |trk,n|
      label = Gtk::Label.new(sprintf('Track %d:', trk.track_no))
      entry = Gtk::Entry.new
      entry.set_text trk.title
      track_entries.push entry
      table.attach label, 0, 1, n, n+1, Gtk::FILL, 0, 0
      table.attach entry, 1, 2, n, n+1, Gtk::FILL | Gtk::EXPAND, 0, 0
      label.show
      entry.show
    }

    scroll = Gtk::ScrolledWindow.new
    scroll.set_usize 300, 300
    scroll.set_policy Gtk::POLICY_NEVER, Gtk::POLICY_AUTOMATIC
    scroll.add_with_viewport table
    table.show

    discard = Gtk::Button.new(' Discard ')
    discard.signal_connect('clicked') { window.destroy }
    apply = Gtk::Button.new(' Save ')

    hbox = Gtk::HBox.new(false, 8)
    pack_end_and_show hbox, discard, false, false, 0
    pack_end_and_show hbox, apply, false, false, 0

    vbox = Gtk::VBox.new(false, 2)
    pack_start_and_show vbox, title_box, false, false, 0
    pack_start_and_show vbox, scroll, true, true, 4
    pack_start_and_show vbox, Gtk::HSeparator.new, false, false, 4
    pack_start_and_show vbox, hbox, false, false, 0

    window.add vbox
    vbox.show

    save_proc = proc {
      disc.title = title_entry.get_text
      disc.artist = artist_entry.get_text
      tracks.each_with_index { |trk,n|
        trk.title = track_entries[n].get_text
      }
      tracks = []
      @playlist.rows.times { |row| tracks.push @playlist.get_row_data(row) }
      list_changed tracks
      if @current then
        track = @playlist.get_row_data(@current)
      else
        track = @playlist.get_row_data(0)
      end
      if track and track.disc == disc then
        @title_label.set_title track.disc.title, track.disc.artist
      end
      @rbcd.save_disc disc, true
      window.destroy
      true
    }

    [ title_entry, artist_entry ].concat(track_entries).each { |w|
      w.signal_connect('activate', &save_proc)
    }
    apply.signal_connect('clicked', &save_proc)

    window.show
  end



  def no_current
    @play_scale.set_sensitive false
    @play_scale.get_adjustment.value = 0.0 if @play_scale.get_adjustment
    if @playlist.rows == 0 then
      @window.set_title ''
      @title_label.clear
      @track_label.clear
    else
      disc = @playlist.get_row_data(0).disc
      @window.set_title disc.title
      @title_label.set_title disc.title, disc.artist
      @track_label.set_track 0, disc.max_track_no
    end
    @time_now_label.clear
    @time_max_label.clear
  end


  public

  def list_changed(tracks)
    current_track = (@current and @playlist.get_row_data(@current))
    @current = nil
    @playlist.freeze
    @playlist.clear
    tracks.each { |track|
      title = track.title
      title = sprintf('(Track %d)', track.track_no) if title.empty?
      t = track.size
      time = sprintf('%02d:%02d.%02d', t.min, t.sec, t.frame)
      n = @playlist.append([ track.disc.slot_no.to_s,
                             track.track_no.to_s, time, title ])
      @playlist.set_row_data n, track
      @playlist.set_track_style n, track
      @current = n if track == current_track
    }
    if @current then
      @playlist.select_row @current, 0
      @playlist.moveto @current, 0, 0, 0
    end
    @playlist.thaw
    @toolbox.set_sensitive((not tracks.empty?))
  end


  def current_changed(track)
    prev = @current
    @current = @playlist.find_row_from_data(track)
    @playlist.unselect_row prev, 0 if prev
    unless track then
      no_current
      @playlist.moveto 0, 0, 0, 0
    else
      raise "no such track in CList: #{track.inspect}" unless @current
      @play_scale.set_sensitive true
      adj = Gtk::Adjustment.new(1.0, 0.0, track.size.to_i, 1.0, 300.0, 0.0)
      adj.signal_connect('value_changed') { |a|
        if @play_scale_changing then
          now = CDROM::Time.new(a.value.to_i)
          now += track.start_addr if @display_abs_time
          @time_now_label.set_time now
        end
      }
      @play_scale.set_adjustment adj
      adj.value = 0.0
      @track_label.set_track track.track_no, track.disc.max_track_no
      if @display_abs_time then
        @time_now_label.set_time track.start_addr
        @time_max_label.set_time track.disc.size
      else
        @time_now_label.clear
        @time_max_label.set_time track.size
      end
      @window.set_title track.disc.title
      @title_label.set_title track.disc.title, track.disc.artist
      @playlist.select_row @current, 0
      @playlist.moveto @current, 0, 0, 0
    end
  end


  def now_playing(track, disc_addr, track_time)
    return if @play_scale_changing
    start = track.start_addr
    if disc_addr >= start then
      @play_scale.get_adjustment.value = disc_addr.to_i - start.to_i
    else
      @play_scale.get_adjustment.value = 0
    end
    if @display_abs_time then
      @time_now_label.set_time disc_addr
    else
      @time_now_label.set_time track_time
    end
  end

end





if $0 == __FILE__ then
  if ARGV.size > 1 then
    print "usage: #{$0} [<device>]\n"
  else
    RbcdGtk.new((ARGV[0] or CDROM::DefaultDevice)).main
  end
end
